class OrderSpec(object):
    def __init__(self, buy=False, sell=False, price=0, qty=0,
                 src_tuple=None, empty=False):
        if src_tuple is not None:
            self.buy, self.sell, self.price, self.qty, self.empty = src_tuple
        else:
            self.buy = buy
            self.sell = sell
            self.price = price
            self.qty = qty
            self.empty = empty

    def to_tuple(self):
        return (self.buy, self.sell, self.price, self.qty, self.empty)

    def __str__(self):
        if self.empty:
            return 'OrderSpec(None)'
        return 'OrderSpec({}, price={}, qty={})'.format(
            'buy' if self.buy else 'sell',
            self.price, self.qty)

    def __repr__(self):
        return self.__str__()


class Order(object):
    def __init__(self, buy=False, sell=False, price=0, qty=0, identifier=0,
                 spec=None, empty=False):
        if spec:
            self.buy = spec.buy
            self.sell = spec.sell
            self.price = spec.price
            self.qty = spec.qty
            self.empty = spec.empty
        else:
            self.buy = buy
            self.sell = sell
            self.price = price
            self.qty = qty
            self.empty = empty
        self.identifier = identifier
        self.processing = True
        self.filled = False
        self.filled_price = None
        self.filled_qty = None

    def fill(self, price, qty):
        self.processing = False
        self.filled = True
        self.filled_price = price
        self.filled_qty = qty

    def reject(self):
        self.processing = False
        self.filled = False

    def gen_spec(self):
        return OrderSpec(self.buy, self.sell, self.price, self.qty)



from .core import PrescientStrategy
from ...order import OrderSpec


class QtyIndependent(PrescientStrategy):
    def __init__(self, qty=0.1):
        super().__init__()
        # a spike is the top of a peak or the bottom of a valley
        self.spikes = []  # (idx, price)
        self.runs = []  # (start_idx, start_price, up?)
        self.run_idx = 0
        self.qty = qty

    def init_pretest(self, fee):
        super().init_pretest(fee)
        self.run_idx = 0

    def read_data(self, tss, prices, qtys):
        super().read_data(tss, prices, qtys)
        uprun = self.prices[1] > self.prices[0]
        downrun = not uprun
        for i, price in enumerate(self.prices[:-1]):
            if uprun and price > self.prices[i+1]:
                uprun = False
                downrun = True
                self.spikes.append((i, price))
            elif downrun and price < self.prices[i+1]:
                downrun = False
                uprun = True
                self.spikes.append((i, price))

        # pprint(self.spikes)

        prev_price = self.spikes[0][1]
        prev_idx = 0
        for idx, price in self.spikes:
            if price > prev_price:
                if (price - prev_price) / prev_price > self.fee:
                    self.runs.append((prev_idx, prev_price, True))  # uprun
                    prev_price = price
                    prev_idx = idx
            elif price < prev_price:
                if (prev_price - price) / prev_price > self.fee:
                    self.runs.append((prev_idx, prev_price, False))  # downrun
                    prev_price = price
                    prev_idx = idx

    def prevaluate(self, idx, ts, price, qty):
        if self.run_idx < len(self.runs):
            start_idx, start_price, run_up = self.runs[self.run_idx]
            if idx + 1 >= start_idx:
                spec = OrderSpec(buy=run_up, sell=not run_up,
                                 price=start_price, qty=self.qty)
                self.run_idx += 1
                return spec
        return OrderSpec(empty=True)



from .core import Strategy
from ..order import Order
from collections import deque


class NaiveStrategyReverse(Strategy):
    '''
        REVERSE basic strategy: buy if we're on a rising price trend, sell if we've been losing money

        Essentially, we're betting against the market trend
    '''
    def __init__(self, n_prices=10, qty=0.01, threshold=0.8, price_inc=0.1):
        super().__init__()
        self.n_prices = n_prices
        self.old_prices = deque(maxlen=n_prices)
        self.qty = qty
        self.threshold = threshold
        self.price_inc = price_inc
        self.uptrend_count = 0
        self.downtrend_count = 0


    def gen_order(self, ts, price, qty, funds, balance):
        order = None
        if len(self.old_prices) > 1:
            if price >= self.old_prices[-1]:
                self.uptrend_count += 1
            if price <= self.old_prices[-1]:
                self.downtrend_count += 1
            if len(self.old_prices) > self.n_prices + 1:
                if self.old_prices[1] >= self.old_prices[0]:
                    self.uptrend_count -= 1
                if self.old_prices[1] <= self.old_prices[0]:
                    self.downtrend_count -= 1
                self.old_prices.popleft()
            
            # reverse Naive strategy: SELL if we're on an uptrend, BUY if we're on a downtrend
            if self.uptrend_count / len(self.old_prices) > self.threshold:
                order = Order(sell=True, price=price + self.price_inc,
                              qty=self.qty, identifier=len(self.orders))
            if self.downtrend_count / len(self.old_prices) > self.threshold:
                order = Order(buy=True, price=price - self.price_inc,
                              qty=self.qty, identifier=len(self.orders))
        self.old_prices.append(price)
        return order



from random import random
from plotly.offline import plot
import plotly.graph_objs as go
from .data import load_data, subarray_with_stride, to_datetimes
import time
import numpy as np
import pprint


def resolve_data(data, fnm, name, datapart,
                 train_prop=0.8, val_prop=0.1):
    '''
        Either read in the dataset or use a caller-provided one
    '''
    if data is not None:
        return data
    train_data, val_data, test_data = load_data(fnm, name,
                                                train_prop, val_prop)
    if datapart == 'train':
        return train_data
    elif datapart == 'val':
        return val_data
    elif datapart == 'test':
        return test_data


def backtest(strategy, initial_funds=1000, initial_balance=0, fill_prob=0.5,
             fee=0.0025, data=None, data_fnm='data/coinbaseUSD.npz',
             history_fnm='histories/backtest.npz',
             data_name='data', datapart='val',
             plot_fnm='temp-plot.html',
             train_prop=0.8, val_prop=0.1, verbose=1, print_freq=10000):

    input_args = locals()  # save the input arguments

    # default to validation
    data = resolve_data(data, data_fnm, data_name, datapart)
    print(data.shape)
    funds = initial_funds  # US dollars

    time1 = time.time()
    fund_history = [funds]
    balance = initial_balance  # bitcoin amounts
    balance_history = [balance]
    initial_worth = initial_funds + initial_balance * data[0][1]
    worth = initial_worth  # total value converted to dollars
    worth_history = [worth]
    balance_worth_history = [balance * data[0][1]]
    ts_history = [data[0][0]]
    strategy.init_backtest(funds, balance, initial_worth)

    # each tuple is the log of an actual transaction at timestamp ts for the
    # listed price/qty
    # we don't have historical data on the order book, so we can only infer
    # what kind of order/sell strategy would have worked based on the actual
    # transactions namely, if we sell at a lower price or buy at a higher price
    # for the same or less quantity, we should succeed
    time2 = time.time()

    for i, (ts, price, qty) in enumerate(data[:-1]):
        if verbose > 1 and i % print_freq == 0:
            print(i, ('worth', worth, 'balance', balance, 'funds', funds))

        next_ts, next_price, next_qty = data[i+1]
        order = strategy.evaluate(ts, price, qty, funds, balance)
        if order is None or order.empty:
            continue

        # in real markets there's a chance that your order won't succeed, say
        # if someone beats you to buying;
        # so we use fill_prob to simulate that "fail chance"

        # this essentially estimates the chance that if we send the same
        # next-historical-transaction, we'll get the deal first
        if random() > fill_prob:
            strategy.reject_order(order.identifier)
            continue

        # try to buy
        if order.buy:
            # we can only guarantee buying if our price is higher than the
            # historical transacted price
            if not order.price >= next_price:
                if verbose > 2:
                    print('price too low')
                strategy.reject_order(order.identifier)
                continue
            else:
                # we can only guarantee a quantity based on historical
                # transactions
                filled_qty = min(order.qty, next_qty)
                filled_price = next_price
                if filled_price * filled_qty > funds:
                    if verbose > 2:
                        print('insufficient funds')
                    strategy.reject_order(order.identifier)
                    continue
                fund_delta = -filled_price * filled_qty
                balance_delta = filled_qty

        # try to sell; same limitations apply as above
        if order.sell:
            if not order.price <= next_price:
                if verbose > 2:
                    print('price too high')
                strategy.reject_order(order.identifier)
                continue
            else:
                filled_qty = min(order.qty, next_qty)
                filled_price = next_price
                if filled_qty > balance:
                    if verbose > 2:
                        print('insufficient balance')
                    strategy.reject_order(order.identifier)
                    continue
                fund_delta = filled_price * filled_qty
                balance_delta = -filled_qty

        # if order succeeded:
        strategy.fill_order(order.identifier, filled_price, filled_qty)
        funds -= filled_price * filled_qty * fee  # pay transaction fee
        funds += fund_delta
        balance += balance_delta

        # historical tracking of what our algorithm predicts
        fund_history.append(funds)
        balance_history.append(balance)
        ts_history.append(next_ts)
        worth = funds + balance * next_price
        worth_history.append(worth)
        balance_worth_history.append(balance * next_price)

    ts_history.append(data[-1][0])
    worth_history.append(worth)
    balance_worth_history.append(balance * next_price)
    fund_history.append(funds)

    time3 = time.time()

    if verbose > 0:
        print('=' * 50)
        print('Backtest summary for:')
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(input_args)
        print('Funds: {} -> {}'.format(initial_funds, funds))
        print('Balance: {} -> {}'.format(initial_balance, balance))
        print('Net worth: {} -> {}'.format(initial_worth, worth))
        buy_hold_worth = (initial_funds / data[0][1]
                          + initial_balance) * data[-1][1]
        print('Buy hold equivelent: {} -> {}'.format(initial_worth,
                                                     buy_hold_worth))

    plot_data = subarray_with_stride(data, 100)

    results = [{'x': to_datetimes(plot_data[:, 0]),
                'y': 1000 * plot_data[:, 1] / plot_data[0, 1],
                'name': 'Buy-hold net worth (Bitcoin Price scaled)',
                'line': dict(width=2.0)},
               {'x': to_datetimes(ts_history),
                'y': balance_worth_history,
                'name': 'Algorithm Nonliquid Bitcoin Worth',
                'fill': 'tozeroy',
                'line': dict(color='rgb(111, 231, 219)')},
               {'x': to_datetimes(ts_history),
                'y': worth_history,
                'name': 'Algorithm Net Worth (Bitcoin + Cash)',
                'fill': 'tonexty',
                'line': dict(width=0.5, color='rgb(184, 247, 212)')}]

    # print(plot_data.shape)
    buy_hold_ts_history = plot_data[:, 0]
    buy_hold_eq_history = 1000 * plot_data[:, 1] / plot_data[0, 1]

    results += strategy.additional_plots()

    if plot_fnm is not None:
        plot_results(results, plot_fnm)

    time4 = time.time()

    if verbose > 0:
        print('Time to load data: {}s'.format(time2 - time1))
        print('Time to train: {}s, {}s per 1000 ticks'.format(
            time3 - time2, (time3 - time2)/len(data)))

        print('Time to plot: {}s'.format(time4-time3))

    np.savez(history_fnm, ts_history=np.asarray(ts_history, dtype='int32'),
             buy_hold_eq_history=buy_hold_eq_history,
             buy_hold_ts_history=np.asarray(buy_hold_ts_history,
                                            dtype='int32'),
             worth_history=worth_history,
             balance_worth_history=balance_worth_history)

    return dict(fund_history=fund_history,
                balance_history=balance_history,
                ts_history=ts_history,
                worth_history=worth_history)


def plot_results(results, plot_name='temp-plot.html'):
    traces = []

    for input_args in results:
        traces.append(go.Scatter(**input_args))

    layout = go.Layout(
        title='Trading performance over time',
        yaxis=dict(
            title='Value (USD)'
        ),
    )
    plot(go.Figure(data=traces, layout=layout), filename=plot_name)



from .backtest import resolve_data
from .data import save_data
import numpy as np


def compute_prescient(strategy, fee=0.0025, data=None,
                      data_fnm='data/coinbaseUSD.npz', data_name='data',
                      datapart='train',
                      history_fnm='prescient_histories/naive.npz',
                      history_name='data'):
    data = resolve_data(data, data_fnm, data_name, datapart)
    order_history = []
    strategy.init_pretest(fee)
    strategy.read_data(tss=data[:, 0], prices=data[:, 1], qtys=data[:, 2])
    for i, (ts, price, qty) in enumerate(data):
        order = strategy.prevaluate(i, ts, price, qty)
        order_history.append(order)
    strategy.read_order_history(order_history)
    order_arr = np.asarray([order.to_tuple() for order in order_history])
    save_data(order_arr, history_fnm, history_name)
    return order_history



import numpy as np
from ..core import Strategy
from ...order import Order, OrderSpec


class PrescientStrategy(Strategy):
    def __init__(self, order_history=None):
        super().__init__()
        self.tss = []
        self.prices = []
        self.qtys = []
        self.order_history = order_history
        self.idx = 0

    def init_pretest(self, fee):
        self.fee = fee

    def init_backtest(self, funds, balance, worth):
        super().init_backtest(funds, balance, worth)
        self.idx = 0

    def read_order_history(self, order_history):
        self.order_history = order_history

    def load_order_history(self, history_fnm, history_name='data'):
        order_arr = np.load(history_fnm)[history_name]
        self.order_history = [OrderSpec(src_tuple=row) for row in order_arr]

    def read_data(self, tss, prices, qtys):
        self.tss = tss
        self.prices = prices
        self.qtys = qtys

    def gen_order(self, ts, price, qty, funds, balance):
        order = Order(spec=self.order_history[self.idx])
        self.idx += 1
        return order



from urllib.request import urlretrieve
import gzip
import os.path
from .data import csv_to_npz

coinbase_usd_url = 'http://api.bitcoincharts.com/v1/csv/coinbaseUSD.csv.gz'
data_dir = 'data'
coinbase_usd_fnm = os.path.join(data_dir, 'coinbaseUSD')
coinbase_usd_csv = coinbase_usd_fnm + '.csv'
coinbase_usd_npz = coinbase_usd_fnm + '.npz'
coinbase_usd_gz = coinbase_usd_csv + '.gz'

if not os.path.exists(coinbase_usd_npz):
    if not os.path.exists('data'):
        os.makedirs('data')

    if not os.path.exists(coinbase_usd_csv):
        print('=' * 50)
        print('Downloading coinabse-usd datset...')
        urlretrieve(coinbase_usd_url, coinbase_usd_gz)

        with gzip.open(coinbase_usd_gz, 'rb') as infile:
            with open(coinbase_usd_csv, 'wb') as outfile:
                for line in infile:
                    outfile.write(line)

    csv_to_npz(coinbase_usd_csv, coinbase_usd_npz)



class Strategy(object):
    '''
        A general strategy framework for deciding when and how much to buy/sell

    '''
    def __init__(self):
        self.orders = []

    def init_backtest(self, funds, balance, worth):
        self.intial_funds = funds
        self.initial_balance = balance
        self.initial_worth = worth

    def evaluate(self, ts, price, qty, funds, balance):
        order = self.gen_order(ts, price, qty, funds, balance)
        if order:
            self.orders.append(order)
        return order

    def fill_order(self, idx, price, qty):
        self.orders[idx].fill(price, qty)

    def reject_order(self, idx):
        self.orders[idx].reject()

    def additional_plots(self):
        return []



import numpy as np
import csv
from datetime import datetime


def to_datetimes(array):
    return [datetime.utcfromtimestamp(x) for x in array]


def from_datetimes(array):
    return [x.timestamp() for x in array]


def subarray_with_stride(array, stride):
    return np.asarray([array[i] for i in range(0, len(array), stride)])


def load_data(fnm, name='data', train_prop=0.8,
              val_prop=0.1):
    data = np.load(fnm)[name]
    train_idx = int(len(data) * train_prop)
    train_data = data[:train_idx]
    val_idx = int(len(data) * (train_prop + val_prop))
    val_data = data[train_idx:val_idx]
    test_data = data[val_idx:]
    return train_data, val_data, test_data


def save_data(data, fnm, name='data'):
    np.savez(fnm, **{name: data})


def csv_to_npz(csv_fnm, npz_fnm, name='data'):
    with open(csv_fnm, 'r') as f:
        reader = csv.reader(f)
        data = np.zeros((sum(1 for row in reader), 3), dtype='float32')
        print(data.shape)
        f.seek(0)
        for i, row in enumerate(reader):
            data[i, :] = int(row[0]), float(row[1]), float(row[2])
    np.savez(npz_fnm, **{name: data})


def load_coinbase_usd():
    return load_data('data/coinbaseUSD.npz', 'data')



from .core import Strategy
from ..order import Order
from collections import deque
import numpy as np
from ..data import to_datetimes, subarray_with_stride


class EMAPriceCrossover(Strategy):
    '''
        We calculate the moving average of the stock, and buy if
        current price is significantly higher and sell if it's significantly
        lower than the moving average. (Price Crossover)

        More information:
        http://www.investopedia.com/articles/active-trading/052014/ \
        how-use-moving-average-buy-stocks.asp
    '''

    def __init__(self, ma_length=100, qty=0.1, decision='increasing',
                 gap_threshold=0.001, trend_threshold=10, verbose=0):
        '''
            If decision == 'increasing', then we will buy (amount = qty) once
            the price line crosses above the EMA and that gap keep increasing
            (similarly for selling)

            If decision == 'once', then we'll only buy up to (amount = qty)
            once every time crossover happens

            qty is the buying/selling limit at any given time.
            Set to -1 for inf

            trend_threshold is the min length a price crossover must last in
            order to be considered a "trend" and not fluxations

            gap_threshold is a fraction for how much above the price must the
            gap be to be considered a cross-over
        '''
        super().__init__()
        self.ma_length = ma_length  # length of Moving Average we care about
        self.alpha = 1/((ma_length+1)/2)  # alpha param in exponential MA
        self.weight_sum = sum((1-self.alpha)**i for i in range(self.ma_length))

        if verbose:
            print('alpha: {}, weight_sum: {}'.format(self.alpha,
                                                     self.weight_sum))

        self.old_prices = deque()
        if qty == -1:
            self.qty = np.inf
        else:
            self.qty = qty
        self.decision = decision
        self.verbose = verbose
        self.identifier = 0
        self.trend_threshold = trend_threshold
        self.gap_threshold = gap_threshold
        self.emas = deque()
        self.ema_tss = deque()

        # diff between last price & last EMA
        # used to track the price gaps for (decision = 'increasing')
        self.last_gap = 0
        self.current_state = 0  # positive: increasing; negative: decreasing
        self.initial_price = None

    def gen_order(self, ts, price, qty, funds, balance):

        if self.initial_price is None:
            self.initial_price = price

        order = None
        qty = min(qty, self.qty)

        if len(self.old_prices) < self.ma_length:
            pass

        elif len(self.emas) == 0:  # we've had the first ma_length data entries
            # get ewma for the first time
            self.emas.append(sum(
                self.old_prices[-1-i] * (1-self.alpha)**i
                for i in range(self.ma_length)) / self.weight_sum)
            self.ema_tss.append(ts)

            if self.verbose > 1:
                print('Init EMA: ', self.emas)
                print('prices: ', self.old_prices)

        else:
            # calculate new EMA
            # we first "unadjust" the weighted sum (by scaling from 1
            # back to the sum of the weights),
            # then we subtract the "expired" price term, shift the weights,
            # and add in the new price term
            ema = ((self.emas[-1] * self.weight_sum -
                    self.old_prices[0] * (1-self.alpha)**(self.ma_length-1))
                   * (1-self.alpha) + price) / self.weight_sum

            if self.verbose > 1:
                if self.verbose > 3 or ts % 1000 == 0:
                    print('ema: ', ema)

            # switching into increasing from decreasing (crossover)
            if self.emas[-1] < self.old_prices[-1] and \
               ema - price >= self.gap_threshold * (ema):

                if self.decision == 'increasing':
                    new_gap = ema - price

                    order = Order(sell=True, price=price,
                                  qty=self.qty, identifier=self.identifier)

                    self.identifier += 1
                    self.last_gap = new_gap
                    self.current_state += 1

                else:  # just order once
                    order = Order(sell=True, price=price,
                                  qty=self.qty, identifier=self.identifier)

            # already increasing
            elif (self.current_state >= self.trend_threshold and
                  self.decision == 'increasing'):
                new_gap = ema - price
                # buy as long as the price gap is increasing
                if new_gap <= self.last_gap:
                    order = Order(sell=True, price=price,
                                  qty=self.qty, identifier=self.identifier)

                    self.identifier += 1
                    self.last_gap = new_gap
                else:
                    self.last_gap = 0
                    self.current_state = 0

            # selling
            elif (self.emas[-1] > self.old_prices[-1] and
                  price - ema >= self.gap_threshold*(ema)):
                if self.verbose > 1:
                    print("crossing over: ema: {}, price: {}"
                          .format(ema, price))

                if self.decision == 'increasing':
                    new_gap = price - ema

                    order = Order(buy=True, price=price,
                                  qty=self.qty, identifier=self.identifier)
                    self.identifier += 1
                    self.last_gap = new_gap
                    self.current_state -= 1

                else:
                    order = Order(buy=True, price=price,
                                  qty=self.qty, identifier=self.identifier)

            elif (self.current_state <= -self.trend_threshold and
                  self.decision == 'increasing'):
                new_gap = price - ema
                # buy as long as the price gap is increasing
                if new_gap >= self.last_gap:
                    order = Order(buy=True, price=price,
                                  qty=self.qty, identifier=self.identifier)
                    self.identifier += 1
                    self.last_gap = new_gap
                else:
                    self.last_gap = 0
                    self.current_state = 0

            self.emas.append(ema)
            self.ema_tss.append(ts)
            self.old_prices.popleft()

        self.old_prices.append(price)
        return order

    def additional_plots(self):
        ema = subarray_with_stride(self.emas, 100)
        tss = to_datetimes(subarray_with_stride(self.ema_tss, 100))
        return [{'x': tss,
                 'y': 1000 * ema / self.initial_price,
                 'name': 'EMA Price (scaled)'}]



import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='cointk',
    version='0.0.10',
    author='Alex Gajewski, Wanqi Zhu, Ashwin Aggarwal',
    author_email='agajews@gmail.com, 1213.ghs@gmail.com, aaggarw99@gmail.com',
    description=('An open-sourced platform for rapid prototyping and testing '
                 'of BitCoin trading strategies, with visualiaztion features '
                 'via our open-sourced iOS app BitBox. Machine Learning '
                 'algorithms coming Soon.'),
    license='MIT',
    url='http://github.com/CoinTK',
    packages=find_packages(),
    long_description=read('README.md'),
    install_requires=['numpy', 'pprint', 'plotly']
)



from .core import PrescientStrategy
from ...order import OrderSpec


class QtySemidependent(PrescientStrategy):
    def __init__(self, qty=0.1):
        super().__init__()
        self.spikes = []  # (idx, price)
        self.spike_idx = 0
        self.qty = qty

    def init_pretest(self, fee):
        super().init_pretest(fee)
        self.spike_idx = 0

    def read_data(self, tss, prices, qtys):
        super().read_data(tss, prices, qtys)
        uprun = self.prices[1] > self.prices[0]
        downrun = not uprun
        for i, price in enumerate(self.prices[:-1]):
            if uprun and price > self.prices[i+1]:
                uprun = False
                downrun = True
                self.spikes.append((i, price))
            elif downrun and price < self.prices[i+1]:
                downrun = False
                uprun = True
                self.spikes.append((i, price))

        # pprint(self.spikes)

    def prevaluate(self, idx, ts, price, qty):
        if self.spike_idx + 2 < len(self.spikes):
            end_idx, end_price = self.spikes[self.spike_idx+1]
            curr_price = self.prices[idx+1]
            if idx > end_idx:
                self.spike_idx += 1
                end_idx, end_price = self.spikes[self.spike_idx+1]
            if end_price > curr_price:
                if (end_price - curr_price) / curr_price > self.fee:
                    return OrderSpec(buy=True, price=curr_price, qty=self.qty)
            if curr_price > end_price:
                if (curr_price - end_price) / curr_price > self.fee:
                    return OrderSpec(sell=True, price=curr_price, qty=self.qty)
        return OrderSpec(empty=True)



from .core import Strategy
from ..order import Order
from collections import deque
import random


class RandomStrategy(Strategy):
    '''
        Random!!
    '''
    def __init__(self, p_buy=0.1, p_sell=0.1, seed=None, randomize_qty=''):
        '''
            randomize_qty = 'Linear' if you want to choose a random amount 
            between (0, qty) every time we buy or sell

            Otherwise it's assumed we buy/sell the maximum amount possible
        '''
        super().__init__()
        self.p_buy = p_buy
        self.p_sell = p_sell
        self.identifier = -1
        self.randomize_qty = randomize_qty

        random.seed(seed)
        


    def gen_order(self, ts, price, qty, funds, balance):
        '''
        randomly buy or sell based on a random number
        '''
        r = random.random()

        if r < self.p_buy + self.p_sell:
            if self.randomize_qty == 'linear':
                qty *= random.random()

            self.identifier += 1
            if r < self.p_buy:
                return Order(buy=True, price=price,
                              qty=qty, identifier=self.identifier)
            else:
                return Order(sell=True, price=price,
                              qty=qty, identifier=self.identifier)
        else:
            return None
